type 'a t = 'a list;;

let hd = function
 | x :: _ -> x
 | _ -> failwith "Lst.hd: empty list"
;;
let tl = function
 | _ :: xs -> xs
 | _ -> failwith "Lst.tl: empty list"
;;
let rec append xs ys = if xs = [] then ys else hd xs :: append (tl xs) ys;;
let rec init i l = if l <= 0 then [] else i :: init i (l - 1);;
let rec take n xs =
 if n <= 0 then [] else match xs with
  | [] -> []
  | x :: xs -> x :: take (n - 1) xs
;;
let rec drop n xs =
 if n <= 0 then xs else match xs with
  | [] -> []
  | _ :: xs -> drop (n - 1) xs
;;
let split_at n xs = (take n xs, drop n xs);;
let rec fold f b = function
 | [] -> b
 | x :: xs -> f x (fold f b xs)
;;
let rec fold_left f b = function
 | [] -> b
 | x :: xs -> fold_left f (f b x) xs
;;
let rec filter p = function
 | [] -> []
 | x :: xs -> if p x then x :: filter p xs else filter p xs
;;
let length xs = fold (fun _ x -> x + 1) 0 xs;;
let map f xs = fold (fun x xs -> f x :: xs) [] xs;;
let rec last = function
 | [x] -> x
 | _ :: xs -> last xs
 | _ -> failwith "Lst.last: empty list"
;;
let rec nth n = function
 | [] -> failwith "Lst.nth: out of bounds"
 | x :: xs -> if n = 0 then x else nth (n - 1) xs
;;
let rec zip2 xs ys = match (xs, ys) with
 | (x :: xs, y :: ys) -> (x, y) :: zip2 xs ys
 | _ -> []
;;
let zip2_with f xs ys = map (fun (x, y) -> f x y) (zip2 xs ys);;
let rec fold1 f = function
 | [x] -> x
 | x :: xs -> f x (fold1 f xs)
 | _ -> failwith "Lst.fold1: empty list"
;;
let rec zip4 ws xs ys zs = match (ws, xs, ys, zs) with
 | (w :: ws, x :: xs, y :: ys, z :: zs) -> (w, x, y, z) :: zip4 ws xs ys zs
 | _ -> []
;;
let rec take_while p = function
 | [] -> []
 | x :: xs -> if p x then x :: take_while p xs else []
;;
let rec drop_while p = function
 | [] -> []
 | x :: xs as ys -> if p x then drop_while p xs else ys
;;
let span p xs = (take_while p xs, drop_while p xs);;
let rev xs =
 let rec rev acc = function
  | [] -> acc
  | x :: xs -> rev (x :: acc) xs
 in
 rev [] xs
;;
let rec until p f x = if p x then x else until p f (f x);;
let concat xs = fold append [] xs;;
